/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "../ftp_main.h"

/* #define SPP_BUFFERSIZE (MIN_VAL(65535, SSIZE_MAX)) */
#define SPP_BUFFERSIZE 65535

struct config_directives_list spp_configuration_directives_array[] = {
	/* distinguish between
	 *   - as transfermode from client
	 *   - as transfermode for server */
	{"enabled",		TAG_STARTUP, "no", EM, WSP },
	{"useforclient",	TAG_STARTUP, "yes", EM, WSP },
	{"useforserver",	TAG_STARTUP, "yes", EM, WSP },
	{ (char*) 0,			  0, (char*) 0, 0, 0 }
};

struct config_namespace_directives_list spp_configuration_directives = {
	"spp",		/* namespace name */
	spp_configuration_directives_array
};

struct config_option_values_list spp_configuration_options_array[] = {
	{"enabled",	{ TRUEFALSE, TERM } },
	{ TERM,         { TERM } }
};

struct config_namespace_options_list spp_configuration_options = {
	"spp",
	spp_configuration_options_array
};

int spp_add_configuration() {
	config_configuration_list_add(&spp_configuration_directives);
	config_options_list_add(&spp_configuration_options);
	return 0;
}

int
transfermode_spp_out(int fd_out,
		     struct transfer_t* transfer,
		     struct proxyinfo* pi) {

	char header[3];
	ssize_t size;
	int error = 0;
	int ret;

	size = transfer->bufsize;
	if (size < 0) {
		jlog(5, "Error reading from buffer in transfermode_spp_out");
		/* send packet saying that there was an error */
		header[0] = '\0';
		header[1] = '\0';
		header[2] = 'E';
		error = -1;
	} else if (size > 0) {
		unsigned char lo = (unsigned int) size % 256;
		unsigned char hi = (unsigned int) (size - lo) / 256;
		header[0] = hi;
		header[1] = lo;
		header[2] = 'D';
	} else if (size == 0) {
		/* send packet saying that this was all the data */
		header[0] = 0;
		header[1] = 0;
		header[2] = 'O';
	}
	jlog(2, "Sending Header to fd %d\n", fd_out);
	ret = writeall(fd_out, &header[0], 3);
	if (ret < 0) {
		jlog(5, "SPP: Error sending packet header to fd %d: %s",
				fd_out, strerror(errno));
		error = -1;
	}
	if (size > 0) {
		ret = writeall(fd_out, transfer->buffer, size);
		if (ret < 0) {
			jlog(5, "SPP: Error sending packet contents to fd %d: %s",
					fd_out, strerror(errno));
			error = -1;
		}
	}
	free_and_zero_char(&transfer->buffer);
	return error;
}

int spp_retr(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_retr("RETR", args, pi, msgbuf, "spp");
}

int spp_list(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_list("LIST", args, pi, msgbuf, "spp");
}

int spp_stor(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_stor("STOR", args, pi, msgbuf, "spp");
}

int transfermode_spp_deinit_server(struct proxyinfo* pi,
				   struct messagebuffer* msgbuf,
				   struct ftp_message *server_msg) {

	int ss = pi->si->connection.serversocket;

	pi->ci->data_connection.dataserversock = -1;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);

	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		return CMD_ABORT;
	}

	pi->ci->data_connection.complete = (server_msg->code == 226);

	ASSUME(server_msg != NULL);
	ASSUME(server_msg->code);

	return TRANSFERMODE_OK;
}

int transfermode_spp_deinit_client(struct proxyinfo* pi,
				   struct messagebuffer* msgbuf,
				   struct ftp_message *server_msg) {

	int cs = pi->si->connection.clientsocket;
	int respcode;

	pi->ci->data_connection.dataclientsock = -1;

	ASSUME(server_msg != NULL);
	if (server_msg->code) {
		ftp_message_strip_respcode(server_msg);
	} else {
		*server_msg = ftp_message_construct(226, "Transfer ok.");
	}

	message_buffer_add(msgbuf, cs,
				server_msg->code, server_msg->msg);
	message_buffer_say_reset(msgbuf, cs, &respcode);
	pi->ci->lastcmd.respcode = respcode;
	ftp_message_free(server_msg);

	return TRANSFERMODE_OK;
}

int transfermode_spp_sorry_client(const char* cmd,
				  const char** args,
				  struct proxyinfo* pi,
				  struct messagebuffer* msgbuf,
				  struct ftp_message *server_msg) {

	pi->ci->data_connection.dataclientsock = -1;

	/* fallback - nothing special here */
	return transfermode_std_sorry_client(cmd, args, pi, msgbuf, server_msg);
}

int transfermode_spp_out_to_client_init(const char* cmd,
					const char** args,
					struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message *server_msg) {

	int cs = pi->si->connection.clientsocket;
	int respcode;

	ASSUME(server_msg != NULL);
	if (server_msg->code) {
		ftp_message_strip_respcode(server_msg);
	} else {
		/* XXX check it */
		*server_msg = ftp_message_construct(150,
				"Opening data connection for transfer.");
	}

	message_buffer_add(msgbuf, cs,
			server_msg->code, server_msg->msg);
	message_buffer_say_reset(msgbuf, cs, &respcode);
	pi->ci->lastcmd.respcode = respcode;

	pi->ci->data_connection.dataclientsock
				= pi->si->connection.clientsocket;

	/* imagine client does SPP to us which does not know
	 * about active or passive modes. Then if we've set
	 * ASCLIENT, we have a problem, so we decide for just one.
	 * Use PASSIVE because it's less complicated normally */
	jlog(7, "clientmode is neither ACTIVE nor PASSIVE but servermode is ASCLIENT. Switch to servermode == PASSIVE");
	pi->ci->data_connection.transfermode.servermode = PASSIVE;

	return TRANSFERMODE_OK;
}

int transfermode_spp_in_from_server_init(const char* cmd,
					 const char** args,
					 struct proxyinfo* pi,
					 struct messagebuffer* msgbuf,
					 struct ftp_message *server_msg) {

	int ss = pi->si->connection.serversocket;
	const char* translated_cmd = cmd;
	int ret;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);

	/* see if we should use SPP */
	if (config_get_bool("spp.enabled", 0) == 0
			|| config_get_bool("spp.useforserver", 0) == 0) {

		jlog(9, "SPP support is not enabled in the configuration");
		return TRANSFERMODE_NOTSUPPORTED;
	}

	/* XXX check SPP Capa */
	ret = capa_has_server_capa("SPP", pi, msgbuf);
	if (ret != TRANSFERMODE_OK) {
		return ret;
	}

	/* translate the command */
	if (strcasecmp(cmd, "LIST") == 0) {
		translated_cmd = "LISTSPP";
	} else if (strcasecmp(cmd, "RETR") == 0) {
		translated_cmd = "RETRSPP";
	} else if (strcasecmp(cmd, "STOR") == 0) {
		translated_cmd = "STORSPP";
	}

	/* is for a LIST, 1 the maximum number of arguments ? */
	if (args[0]) {
		sayf(ss, "%s %s\r\n", translated_cmd, args[0]);
	} else {
		sayf(ss, "%s\r\n", translated_cmd);
	}
	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		return TRANSFERMODE_FATAL;
	}

	if (server_msg->code != 150) {
		jlog(4, "Server returned invalid response after RETRSPP: %s",
							server_msg->msg);
		return TRANSFERMODE_FAIL;
	}

	pi->ci->data_connection.dataserversock
				= pi->si->connection.serversocket;

	return TRANSFERMODE_OK;
}

int transfermode_spp_out_to_server_init(const char* cmd,
					const char** args,
					struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message *server_msg) {

	return transfermode_spp_in_from_server_init(cmd, args,
						pi, msgbuf, server_msg);
}

int transfermode_spp_in_from_client_init(const char* cmd,
					 const char** args,
					 struct proxyinfo* pi,
					 struct messagebuffer* msgbuf,
					 struct ftp_message *server_msg) {

	return transfermode_spp_out_to_client_init(cmd, args,
						pi, msgbuf, server_msg);
}

int
transfermode_spp_in(int fd_in,
		    struct transfer_t* transfer,
		    struct proxyinfo* pi) {
	char header[3];
	size_t size = 0;
	int error = 0;
	int ret;
	size_t header_size = 3;

	jlog(2, "Reading from fd %d", fd_in);
	ret = readall(fd_in, header, &header_size);
	if (ret < 0) {
		jlog(5, "SPP: Error reading header from fd %d: %s",
				fd_in, strerror(errno));
		error = -1;
	} else {
		size = (size_t) ((unsigned char) header[0]) * 256
				+ (size_t) ((unsigned char) header[1]);
		if (size == 0 && header[2] == 'E') {
			transfer->bufsize = 0;
			transfer->status = TRANSFER_STATUS_ERROR;
			error = -1;
		} else if (size == 0 && header[2] == 'O') {
			/* fine */
			transfer->bufsize = 0;
			transfer->status = TRANSFER_STATUS_FINISHED;
		} else {
			/* data transfer */
			transfer->buffer = (char*) malloc(size);
			enough_mem(transfer->buffer);
			transfer->bufsize = size;
			transfer->status = TRANSFER_STATUS_NORMAL;
			ret = readall(fd_in, transfer->buffer, &size);
			if (ret < 0) {
				error = -1;
			} else {
				jlog(2, "Got buffer from fd %d (size %d)",
							fd_in, size);
			}
		}
	}
	return error;
}

/*  - TEST
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


int main(void) {
	int fd[2];
	pipe(fd);
	if (fork() == 0) {
		int fd_in = open("/temp/test.pl", 0);
		close(fd[0]);
		spp_send(fd_in, fd[1]);
		close(fd_in);
		close(fd[1]);
	} else {
		int fd_out;
		unlink("/tmp/copy");
		fd_out = open("/tmp/copy", O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
		close(fd[1]);
		spp_receive(fd[0], fd_out);
		close(fd_out);
		close(fd[0]);
	}
	return 0;
}

void enough_mem(const void* p) { return; }

void jlog(int level, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	fflush(stderr);
	va_end(args);
	return;
}

*/

/*
int ftp_spp_retr_init(const char* cmd, const char** args,
		struct proxyinfo* pi, struct messagebuffer* msgbuf) {

	int cs = pi->si->connection.clientsocket;
	int ss = pi->si->connection.serversocket;
	struct clientinfo* ci = pi->ci;
	struct ftp_message msg;
	int servermode = ci->data_connection.transfermode.servermode;
	int ret;

	if (servermode == PASSIVE || servermode == ASCLIENT) {
		ret = pasvserver(pi, msgbuf);
	} else {
		ret = activeserver(pi, msgbuf, 0);
	}

	ci->lastcmd.filename = args[0];
	sayf(ss, "RETR %s\r\n", ci->lastcmd.filename);

	msg = ftp_readall(ss);
	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		return CMD_ABORT;
	}

	if (msg.code != 150 && msg.code != 125) {
		jlog(4, "Server returned invalid response: %s", msg.msg);
		ftp_message_free(&msg);
		return CMD_ERROR;
	}
	ftp_message_free(&msg);

	message_buffer_add(msgbuf, cs,
			150, "RETRSPP understood\r\n");
	message_buffer_say_reset(msgbuf, cs);

	ftp_transfer(pi, msgbuf,
			&(pi->ci->data_connection.dataserversock),
			&(pi->si->connection.clientsocket));
	return CMD_HANDLED;
}
*/
/*
int ftp_spp_retr_send(const char* cmd, const char** args,
		struct proxyinfo* pi, struct messagebuffer* msgbuf) {

	int cs = pi->si->connection.clientsocket;
	int ss = pi->si->connection.serversocket;
	message_buffer_add(msgbuf, cs,
			200, "Sending RETRSPP\r\n");

	return CMD_HANDLED;
}
*/

int spp_config_init() {
	config_register_hook(0, spp_add_configuration);
	return 0;
}

int spp_commandhandler_init() {
	if (config_get_bool("spp.enabled", 0) == 1
			&& config_get_bool("spp.useforclient", 0) == 1) {

		capa_add("SPP");

		cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
				CMD_SERIALIZE_REMOVE_ONE,
				"RETRSPP", "SPP RETR", 1, 1,
				-1,
				spp_retr,
				(char*) 0, (cmd_hookfunc_t) 0);

		cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
				CMD_SERIALIZE_REMOVE_ONE,
				"LISTSPP", "SPP LIST", 0, 1,
				-1,
				spp_list,
				(char*) 0, (cmd_hookfunc_t) 0);

		cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
				CMD_SERIALIZE_REMOVE_ONE,
				"STORSPP", "SPP STORE", 1, 1,
				-1,
				spp_stor,
				(char*) 0, (cmd_hookfunc_t) 0);
	}

	transfermode_register(150,
			TRANSFERMODE_RECV_SERVER,
			"spp",
			transfermode_spp_in,
			transfermode_spp_in_from_server_init,
			transfermode_spp_deinit_server,
			NULL,
			"SPP IN from server");

	transfermode_register(150,
			TRANSFERMODE_SEND_CLIENT,
			"spp",
			transfermode_spp_out,
			transfermode_spp_out_to_client_init,
			transfermode_spp_deinit_client,
			transfermode_spp_sorry_client,
			"SPP OUT to client");

	transfermode_register(150,
			TRANSFERMODE_RECV_CLIENT,
			"spp",
			transfermode_spp_in,
			transfermode_spp_in_from_client_init,
			transfermode_spp_deinit_client,
			transfermode_spp_sorry_client,
			"SPP IN from client");

	transfermode_register(150,
			TRANSFERMODE_SEND_SERVER,
			"spp",
			transfermode_spp_out,
			transfermode_spp_out_to_server_init,
			transfermode_spp_deinit_server,
			NULL,
			"SPP OUT to server");

	return 0;
}

int spp_init() {
	module_hook_register(MODULE_HOOK_PROXY_INIT_CONFIG,
			     spp_config_init,
			     "spp config init");

	module_hook_register(MODULE_HOOK_FTP_CMDHANDLER_INIT,
			     spp_commandhandler_init,
			     "spp commandhandler init");
	return 0;
}
