/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <ctype.h>
#include <sys/stat.h>
#include "jproxygw.h"

#ifdef HAVE_LIBWRAP
#include <syslog.h>
#include <tcpd.h>
#ifndef LIBWRAP_ALLOW_FACILITY
#	define LIBWRAP_ALLOW_FACILITY LOG_AUTH
#endif
#ifndef LIBWRAP_ALLOW_SEVERITY
#	define LIBWRAP_ALLOW_SEVERITY LOG_INFO
#endif
#ifndef LIBWRAP_DENY_FACILITY
#	define LIBWRAP_DENY_FACILITY LOG_AUTH
#endif
#ifndef LIBWRAP_DENY_SEVERITY
#	define LIBWRAP_DENY_SEVERITY LOG_WARNING
#endif
int allow_severity = LIBWRAP_ALLOW_FACILITY | LIBWRAP_ALLOW_SEVERITY;
int deny_severity  = LIBWRAP_DENY_FACILITY  | LIBWRAP_DENY_SEVERITY;
#endif

extern int chlds_exited;
extern int should_read_config;
extern int should_terminate;


/* bindport binds to the specified PORT on HOSTNAME (which may also be a
 * dot-notation IP and returns the socket descriptor
 *
 * Parameters: hostname & port: Where to bind
 *
 * Return value: The socket descriptor of the bound socket
 *
 * Called by: waitclient
 *
 * */

int bindport(const char *hostname, int port, struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;

	unsigned long inetaddr = 1;
	int shandle;
	int one = 1;
	struct sockaddr_in sin;
	unsigned long host_ip;

	memset((void*)&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	inetaddr = inet_addr(hostname);
	if (inetaddr == (unsigned long int) UINT_MAX) {
		/* see if HOSTNAME is an interface */
		if (get_interface_ip(hostname, &sin) == 0) {
			/* found it */
			/* sin.sin_addr is already set */
			sin.sin_family = AF_INET;
		} else {
			/* HOSTNAME was probably a name since inet_addr
			 * returned an error.
			 * Look up the name to get the IP
			 */
			host_ip = hostent_get_ip(&si->hostcache, hostname);
			if (host_ip == (unsigned long int) UINT_MAX) {
				jlog(1, "Could not resolve %s: %s",
					hostname, strerror(errno));
				perror("Could not resolve the hostname");
				return -1;
			}
			sin.sin_addr.s_addr = host_ip;
		}
	}
	else {
		/* okay, HOSTNAME was a valid dot-notation IP */
		sin.sin_addr.s_addr = inetaddr;
	}
	sin.sin_port = htons(port);

	/* become root again - use our function instead of plain
	 * setuid() for the logging */
	if (changeid(PRIV, UID, "Changing ID to root (socket(), bind())", pi)) {
		return -1;
	}

	shandle = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (shandle < 0) {
		jlog(1, "Error creating socket to bind: %s", strerror(errno));
		perror("Error creating socket to bind to");
		return -1;
	}
	if (setsockopt(shandle, SOL_SOCKET, SO_REUSEADDR,
				(void*) &one, sizeof(one)) < 0) {
		jlog(3, "Error setting socket to SO_REUSEADDR");
	}
	if (bind(shandle, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		jlog(1, "Error binding: %s", strerror(errno));
		perror("Error binding");
		return -1;
	}
	if (listen(shandle, 5) < 0) {
		jlog(1, "Error listening on the socket: %s", strerror(errno));
		perror("Error listening on a bound socket");
		return -1;
	}

	jlog(6, "Listening on %s, port %d, fd %d", hostname, port, shandle);

	if (changeid(UNPRIV, EUID,
			"Changing id back (socket(), bind())", pi) < 0) {
		return -1;
	}

	return shandle;
}

int unlisten_on_ifaces(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;
	int i;

	for (i = 0; i <= si->connection.listen_set.maxfd; i++) {
		if (FD_ISSET(i, &si->connection.listen_set.set)) {
			jlog(8, "Closing fd %d", i);
			close(i);
		}
	}
	FD_ZERO(&si->connection.listen_set.set);
	si->connection.listen_set.maxfd = 0;
	free_and_zero_char(&si->connection.listen_specs);
	return 0;
}

int listen_on_ifaces(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;

	char* part =0, *portdel =0;
	const char* hostnames;
	char* hostnames2;
	size_t hostnames2size;
	int offset, port, shandle, i;

	offset = 0;
	si->connection.boundsocket_niface = 0;
	FD_ZERO(&si->connection.listen_set.set);
	si->connection.listen_set.maxfd = 0;
	free_and_zero_char(&si->connection.listen_specs);

	hostnames = config_get_option("listen");
	ASSUME(hostnames != NULL);
	/* we have to make the string suitable for quotstrtok by appending a
	 * WHITESPACE character */
	hostnames2size = strlen(hostnames) + 2;
	hostnames2 = (char*) malloc(hostnames2size);
	enough_mem(hostnames2);
	snprintf(hostnames2, hostnames2size, "%s ", hostnames);

	while ((part = quotstrtok(hostnames2, WHITESPACES, &offset))) {
		/* do some counting */
		si->connection.boundsocket_niface++;
		free(part);
	}
	si->connection.boundsocket_list =
		(int*) malloc(sizeof(int) * si->connection.boundsocket_niface);
	enough_mem(si->connection.boundsocket_list);

	offset = 0; i = 0;
	while ((part = quotstrtok(hostnames2, WHITESPACES, &offset))) {
		portdel = strchr(part, ':');
		if (!portdel) {
			jlog(3, "Invalid IP/Port specification: %s", part);
			free(part);
			continue;
		}
		*portdel = (char) 0;
		errno = 0;
		port = strtol(portdel+1, (char**) 0, 10);
		if (errno || port > 65535 || port <= 0) {
			jlog(4, "Invalid port specification: %s", portdel);
			FD_ZERO(&si->connection.listen_set.set);
			si->connection.listen_set.maxfd = -1;
			return -1;
		}
		portdel = (char*) 0;

		jlog(9, "binding to %s, port %d", part, port);

		shandle = bindport(part, port, pi);
		free(part);
		part = (char*) 0;
		if (shandle < 0) {
			free(hostnames2);
			FD_ZERO(&si->connection.listen_set.set);
			si->connection.listen_set.maxfd = -1;
			return -1;
		}
		FD_SET(shandle, &si->connection.listen_set.set);
		si->connection.listen_set.maxfd
			= MAX_VAL(si->connection.listen_set.maxfd, shandle);
		si->connection.boundsocket_list[ i++ ] = shandle;
	}
	free(hostnames2);
	hostnames2 = (char*) 0;

	if (si->connection.boundsocket_niface == 0) {
		jlog(2, "No interfaces found to bind to");
		FD_ZERO(&si->connection.listen_set.set);
		si->connection.listen_set.maxfd = -1;
		return -1;
	}

	si->connection.listen_specs = strdup(hostnames);
	enough_mem(si->connection.listen_specs);
	return 0;
}


static
int get_connecting_socket(struct proxyinfo* pi) {
#ifdef HAVE_SOCKLEN_T
	socklen_t size;
#else
	int size;
#endif
	static int nfd;
	int shandle, ahandle;
	struct sockaddr_in sin;
	struct serverinfo* si = pi->si;
	struct descriptor_set d_set = si->connection.listen_set;

	size = sizeof(sin);

	/* is there no remaining ready fd from the last select() ? */
	if (nfd == 0) {
		while (1) {
			/* eternal select() */
			/* nfd returns the number of fds that are ready */
			nfd = select(d_set.maxfd + 1, &d_set.set, 0, 0, 0);
			if (nfd > 0) {
				break;
			}
			if (errno == EINTR) {
				if (chlds_exited > 0) {
					get_chld_pid(pi);
				}
				if (should_read_config) {
					jlog(9, "Re-reading configuration");
					if (reread_config(pi) < 0) {
						return -1;
					}
					/* restore / re-init the descriptor
					 * set. It might have changed after
					 * reread_config() */
					d_set = si->connection.listen_set;
				}
				if (should_terminate) {
					jlog(9, "Terminating...");
					return -1;
				}
				continue;
			}
			jlog(1, "select() failed: %s, nfd: %d", strerror(errno), nfd);
			return -1;
		}
	}
	/* a descriptor is ready */
	shandle = 0;
	while (!FD_ISSET(shandle, &d_set.set)) {
		shandle++;
	}
	/* one descriptor less is ready in the set */
	nfd--;

	while(1) {
		ahandle = accept(shandle, (struct sockaddr *) &sin, &size);
		if (ahandle < 0) {
			switch(errno) {
				case EINTR:
				    /* signal */
				case ECONNRESET:
				    /* client quit before three-way-handshake */
				case ENETDOWN:
				     /* the others: see manpage */
#ifdef EPROTO
				case EPROTO:
#endif
				case ENOPROTOOPT:
				case EHOSTDOWN:
#ifdef ENONET
				case ENONET:
#endif
				case EHOSTUNREACH:
				case EOPNOTSUPP:
				case ENETUNREACH:
					continue;
			}
			jlog(1, "accept() failed: %s", strerror(errno));
			return -1;
		}
		break;
	}
	return ahandle;
}


static
int client_setup(struct proxyinfo* pi) {
	struct sigaction sa;
	int cs = pi->si->connection.clientsocket;
	struct serverinfo* si = pi->si;
	int config_request_transparent;
	int proxy_loop_detected;
	unsigned long int peer_ip = get_uint_peer_ip(cs);
	struct sockaddr_in t_in;
	struct sockaddr_in c_in;
	const char* client_name;
	int ret;
#ifdef HAVE_LIBWRAP
	struct request_info req;
	int libwrap_allow;
#endif

	if (peer_ip == (unsigned long int) UINT_MAX) {
		return -1;
	}

	/* The clients ignore the SIGHUP signal. Thus
	 * the user can issue a killall -HUP jftpgw
	 * and the master jftpgw process rereads its
	 * configuration file without affecting the
	 * child servers */

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction(SIGHUP, &sa, 0);

	/* And they just reap the status of exited children */
	sa.sa_handler = reap_chld_info;
	sigemptyset(&sa.sa_mask);
#ifndef WIN32
	sa.sa_flags = SA_RESTART;
#endif
	sigaction(SIGCHLD, &sa, 0);

	jlog(7, "Connection from %s", get_char_peer_ip(cs));
	c_in = socketinfo_get_local_sin(cs);
	jlog(7, "Client tried to connect to %s on port %d",
					inet_ntoa(c_in.sin_addr),
					ntohs(c_in.sin_port));

	si->connection.addr_to_client = c_in.sin_addr.s_addr;
	client_name = hostent_get_name(&si->hostcache,
					si->connection.addr_to_client);
	if (client_name) {
		si->connection.client_name = strdup(client_name);
	} else {
		si->connection.client_name = strdup(inet_ntoa(c_in.sin_addr));
	}
	enough_mem(si->connection.client_name);
	si->connection.proxy_ip = c_in.sin_addr.s_addr;
	si->connection.proxy_port = ntohs(c_in.sin_port);

	/* see if we are in transparent mode */
	t_in = socketinfo_get_transparent_target_sin(cs);
	jlog(7, "Transparent target seems to be %s on port %d",
					inet_ntoa(t_in.sin_addr),
					ntohs(t_in.sin_port));

	jlog(9, "Checking TAG_GLOBAL | TAG_FROM | TAG_PROXYIP | TAG_PROXYPORT | TAG_TIME | TAG_SERVERTYPE");
	ret = config_shrink_config(peer_ip,	/* source IP */
				-1,		/* dest IP */
				(char*) 0,	/* dest name */
				0,		/* dest port */
				(char*) 0,	/* dest user */
				-1,		/* forwarded IP */
				(char*) 0,	/* forwarded destination */
				0,		/* forwarded destinationport */
				(char*) 0,	/* forwarded username */
				0,		/* set no specific time */
						/* ip of proxy if   */
				si->connection.proxy_ip,
						/* port of proxy if */
				si->connection.proxy_port,
						/* global variable */
				si->servertype,
				si->ready_to_serve,
				&si->hostcache,
				TAG_CONNECTED
				);
	if (ret != 0) {
		jlog(2, "Error shrinking config data");
		return ret;
	}

#ifndef HAVE_LIBWRAP
	si->tcp_wrapper = 0;
#endif

	if (si->tcp_wrapper) {
#ifdef HAVE_LIBWRAP
		/* use libwrap(tcp_wrapper) for access control
		 *   patch originally by <fukachan@fml.org> but modified
		 *   since then.
		 * It is useful to check both keywords "jftpgw" and "ftp-gw"
		 * for TIS Gauntlet flavour.
		 */
		const char* servicenames_opt
				= config_get_option("libwrapservicenames");
		struct slist_t *servicenames, *servicenames_cur;

		ASSUME(servicenames_opt != NULL);
		servicenames = config_split_line(servicenames_opt,
								WHITESPACES);
		servicenames_cur = servicenames;

		libwrap_allow = 0;
		while (servicenames_cur) {
			/* XXX request_set would give a better performance */
			request_init(&req,
				RQ_DAEMON, servicenames_cur->value,
				RQ_CLIENT_ADDR, conv_ip2char(peer_ip),
				NULL);
			libwrap_allow = hosts_access(&req);
			if (libwrap_allow > 0) {
				break;
			}
			servicenames_cur = servicenames_cur->next;
		}
		slist_destroy_free(servicenames);
		servicenames = servicenames_cur = (struct slist_t*) 0;

		if (libwrap_allow) {
			jlog(5, "%s is allowed to connect (libwrap).", 
					conv_ip2char(peer_ip));
		} else {
			say(cs, "500 You are not allowed to "
					"connect. Goodbye.\r\n");
			jlog(5, "%s was not allowed to connect (libwrap).",
					conv_ip2char(peer_ip));
			close(cs);
			return -2;
		}
#endif
	} else {
		if (config_compare_option("access", "allow") == 0) {
			say(cs, "500 You are not allowed to "
				"connect. Goodbye.\r\n");
			jlog(5, "%s was not allowed to connect.",
						conv_ip2char(peer_ip));
			close(cs);
			return -2;
		} else {
			jlog(6, "%s is allowed to connect.",
					conv_ip2char(peer_ip));
		}
	}

	config_request_transparent =
		config_get_bool("transparent-proxy", 0)
		|| config_get_option("transparent-forward") ?  1 : 0;

	/* see if the proxy loops. It loops when the source has the
	 * same IP and port as the destination. There is no need for such a
	 * configuration "in the wild". Is there? */

	jlog(9, "Proxy loop check. c_in_ip: %s, t_in_port: %d, "
		"c_in_port: %d", inet_ntoa(c_in.sin_addr),
		ntohs(t_in.sin_port), ntohs(c_in.sin_port));

	if (get_uint_peer_ip(cs) == t_in.sin_addr.s_addr
		&& t_in.sin_addr.s_addr == c_in.sin_addr.s_addr
		&& t_in.sin_port == c_in.sin_port) {

		jlog(6, "Proxy loop detected! Disabling transparent proxy");
		proxy_loop_detected = 1;
		config_option_list_delete("transparent-proxy");
		config_request_transparent = 0;
	} else {
		jlog(8, "No proxy loop detected.");
		proxy_loop_detected = 0;
	}

	if (config_request_transparent) {
		jlog(9, "Enabling transparent proxy");
		si->connection.transparent = TRANSPARENT_YES;
		si->connection.transparent_destination = t_in;

		/* Determine the IP the client sees of us */
		if (config_compare_option("getinternalip", "configuration")) {
			const char* address
				= config_get_option("dataclientaddress");
			if (!address) {
				jlog(3, "Option \"getinternalip\" is set to \"configuration\", but no address specified (option \"dataclientaddress\"). Falling back to INADDR_ANY.");
				si->connection.addr_to_client = INADDR_ANY;
			} else {
				si->connection.addr_to_client =
					config_get_addroption("dataclientaddress", INADDR_ANY);
			}
		} else {
			si->connection.addr_to_client =
				socketinfo_get_local_addr_by_sending(cs, pi);
		}
	} else {
		jlog(9, "No transparent proxy support");
		si->connection.transparent = TRANSPARENT_NO;
		si->connection.transparent_destination.sin_port = 0;
		si->connection.transparent_destination.sin_addr.s_addr = 0;
	}

	/* init per-connection logfiles */
	if (log_init_connect(pi) < 0) {
		return -1;
	}

	/* seed the random number generator */
	srand(time(NULL));

	return 0;
}


/* waitclient() waits for a client to connect. It binds to the ports and
 * listens to them. If a connection comes in, jftpgw forks. The parent process
 * keeps on listening whereas the child process handles the connection
 */

int waitclient(struct proxyinfo* pi, struct protocolhandler* protocol) {

	struct serverinfo* si = pi->si;

	int i, ahandle, chldpid =0;
	const char* option = 0;
	struct sigaction sa;
	unsigned int peer_ip;
	struct sockaddr_in c_in;
	time_t now;

	if (si->multithread) {
		daemonize();
	}

	if (changeid(UNPRIV, EUID,
			"Changing id back (socket(), bind())", pi) < 0) {
		return -1;
	}

	if (listen_on_ifaces(pi) < 0) {
		jlog(9, "Error binding... Aborting");
		return -1;
	}

	/* we have successfully bound */

	/* become root again - use our function instead of plain
	 * setuid() for the logging */
	if (changeid(PRIV, UID, "Changing ID to root (pidfile)", pi) < 0) {
		return -1;
	}

	option = config_get_option("pidfile");
	if (option) {
		FILE* pidf;
		umask(022);
		pidf = fopen(option, "w");
		if (pidf) {
			fprintf(pidf, "%ld\n", (long) getpid());
			fclose(pidf);
			/* if successful register function to remove the
			 * pidfile */
			si->pidfile_created = 1;
		} else {
			jlog(2, "Error creating pidfile %s", option);
		}
	}

	/* this has to be done for the daemonization. We do it now after
	 * the pidfile has been created */
	umask(0);

	si->ready_to_serve = SVR_LAUNCH_READY;

	if (stage_action("startsetup", pi) < 0) {
		return -1;
	}

	sa.sa_handler = childterm;
	chlds_exited = 0;
	sigemptyset (&sa.sa_mask);
#ifndef WIN32
	sa.sa_flags = SA_RESTART;
#endif
	sigaction (SIGCHLD, &sa, 0);


	/* Close stdin,stdout,stderr */
	for(i = 0; i <= 2 && si->multithread; i++) {
		close(i);
	}
	si->main_server_pid = getpid();
	atexit(sayterminating);

	while(1) {
		ahandle = get_connecting_socket(pi);
		if (ahandle == -1) {
			/* either select() or accept() failed */
			/* I don't try resume here because we are in an
			 * endless loop. The danger of the programm falling
			 * into an infinite loop consuming all cpu time is
			 * too big... */
			jlog(8, "get_connecting_socket() returned error code");
			return -1;
		}

		c_in = socketinfo_get_local_sin(ahandle);
		peer_ip = get_uint_peer_ip(ahandle);
		now = time(NULL);
		config_counter_increase(peer_ip,               /* from ip */
					c_in.sin_addr.s_addr,  /* proxy_ip */
					ntohs(c_in.sin_port),  /* proxy_port */
					now,                /* specific_time */
					si->servertype,
					&(si->hostcache));
		si->connection.clientsocket = ahandle;

		if (config_check_limit_violation()) {
			/* protocol dependent reject */
			(protocol->proto_reject)(pi);
			config_counter_decrease(peer_ip,       /* from ip */
					c_in.sin_addr.s_addr,  /* proxy_ip */
					ntohs(c_in.sin_port),  /* proxy_port */
					now,                /* specific_time */
					si->servertype,
					&(si->hostcache));
			continue;
		}
		if (si->multithread) {
			if ((chldpid = fork()) < 0) {
				jlog(1, "Error forking: %s", strerror(errno));
				close(ahandle);
				return -1;
			}
			if (chldpid > 0) {
				/* parent process */
				/* register the PID */
				register_pid(chldpid, peer_ip,
					c_in.sin_addr.s_addr,    /* proxy_ip */
					ntohs(c_in.sin_port),  /* proxy_port */
					now,                /* specific_time */
					pi);
				close(ahandle);
			}
			if (chldpid == 0) {
				/* child process */
				jlog(8, "forked to pid %d", getpid());
			}
		}
		if (!si->multithread || chldpid == 0) {
			for (i = 0; i < si->connection.boundsocket_niface; i++) {
				close(si->connection.boundsocket_list[i]);
			}
			free(si->connection.boundsocket_list);
			si->connection.boundsocket_list = (int*) 0;

			if ((protocol->proto_connect)(pi) < 0) {
				return -1;
			}
			if (client_setup(pi) < 0) {
				say(si->connection.clientsocket,
				      "421 Error setting up (see logfile)\r\n");
				return -1;
			}
			/* protocol dependent accept */
			return (protocol->proto_accept)(pi);
		}
	}
}

int inetd_connected(int sock,
		    struct proxyinfo* pi,
		    struct protocolhandler* protocol) {
	if (stage_action("startsetup", pi) < 0) {
		return -1;
	}
	pi->si->connection.clientsocket = sock;
	jlog(8, "Handling inetd connection");
	if ((protocol->proto_connect)(pi) < 0) {
		return -1;
	}
	if (client_setup(pi) < 0) {
		return -1;
	}
	/* protocol dependent accept */
	return (protocol->proto_accept)(pi);
}


