/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "jproxygw.h"

#define FILTER_MAX_FILTERS		10

static struct {
	struct filter_t filter[FILTER_MAX_FILTERS];
	int nfilters;
} filters;

void filter_init(void) {
	filters.nfilters = 0;
	memset(&filters, 0, sizeof(filters));

	/* XXX should it stay here ? */
	filter_register("throughput", FILTER_TYPE_UPSTREAM, filter_limit_bandwidth,
			1,
			"limit upstream bandwidth to x kb/s");
	filter_register("throughput", FILTER_TYPE_DOWNSTREAM, filter_limit_bandwidth,
			1,
			"limit downstream bandwidth to x kb/s");
}

void filter_deinit(void) {
	int i;
	for (i = 0; i < filters.nfilters; i++) {
		free(filters.filter[i].name);
		free(filters.filter[i].description);
	}
	filters.nfilters = 0;
	memset(&filters, 0, sizeof(filters));
}

static
int filter_compare(const void *a, const void* b) {
	struct filter_t* f_a = (struct filter_t*) a;
	struct filter_t* f_b = (struct filter_t*) b;
	
	if (f_a->rank == f_b->rank) {
		return 0;
	}

	if (f_a->rank < f_b->rank) {
		return -1;
	}

	return 1;
}

static
void filter_set_rank(const char* name, int type, int rank) {
	int i;
	int found = 0;
	for (i = 0; i < filters.nfilters; i++) {
		if (filters.filter[i].type == type
			&& strcmp(filters.filter[i].name, name) == 0) {

			filters.filter[i].rank = rank;
			found = 1;
			break;
		}
	}
	if (found == 0) {
		jlog(4, "Filter with name \"%s\" not found (or not enabled). Check your filterorderupstream and filterorderdownstream options", name);
	}
}

static
void filter_apply_ranks() {
	const char* order_downstream = config_get_option("filterorderdownstream");
	const char* order_upstream = config_get_option("filterorderupstream");

	struct slist_t* order_d = (struct slist_t*) 0;
	struct slist_t* order_u = (struct slist_t*) 0;

	struct slist_t* p;
	int rank = 0;

	if (order_downstream) {
		order_d = config_split_line(order_downstream, WHITESPACES);
	}
	if (order_upstream) {
		order_u = config_split_line(order_upstream, WHITESPACES);
	}

	p = order_d;
	while(p && p->value) {
		filter_set_rank(p->value, FILTER_TYPE_DOWNSTREAM, rank);
		rank++;
		p = p->next;
	}

	p = order_u;
	while(p && p->value) {
		filter_set_rank(p->value, FILTER_TYPE_UPSTREAM, rank);
		rank++;
		p = p->next;
	}

	slist_destroy_free(order_d);
	slist_destroy_free(order_u);
}

void filters_sort() {
	filter_apply_ranks();
	/* sort the array with ascending ranks */
	qsort(&filters.filter[0], filters.nfilters,
			sizeof(struct filter_t), filter_compare);
}

int filter_register(const char* name,
		     int type,
		     filter_func_t func,
		     int active,
		     const char* description) {

	if (filters.nfilters == FILTER_MAX_FILTERS) {
		jlog(5, "no more space for filters");
		return -1;
	}

	/* nfilters contains the index of the next free slot */

	/* XXX check name - no spaces for example... */
	filters.filter[filters.nfilters].name = strdup(name);
	enough_mem(filters.filter[filters.nfilters].name);

	filters.filter[filters.nfilters].filter_func = func;
	filters.filter[filters.nfilters].type = type;
	filters.filter[filters.nfilters].active = active;
	/* if they are not listed in the config options for the order, they
	 * will get a higher number and will be sorted at the end. */
	filters.filter[filters.nfilters].rank = FILTER_MAX_FILTERS;

	filters.filter[filters.nfilters].description = strdup(description);
	enough_mem(filters.filter[filters.nfilters].description);

	filters.nfilters++;

	filters_sort();

	return 0;
}

/*
static
int filter_delete(int i) {
	int j;

	* delete the filter at position i and pull all the following entries
	 * one entry down *
	free(filters.filter[i].name);
	free(filters.filter[i].description);

	for (j = i; j < filters.nfilters; j++) {
		i++; * i is one ahead of j now *
		* copy filter in i to j *
		filters.filter[j] = filters.filter[i];
	}

	filters.nfilters--;

	* that's just cosmetics *
	filters.filter[filters.nfilters].name = (char*) 0;
	filters.filter[filters.nfilters].description = (char*) 0;
	filters.filter[filters.nfilters].type = 0;
	filters.filter[filters.nfilters].filter_func = (filter_func_t) 0;

	return 0;
}

int filter_unregister(const char* name, int type) {
	int i;
	int ret = -1;

	for (i = 0; i < filters.nfilters; i++) {
		if (filters.filter[i].type == type
			&& strcmp(filters.filter[i].name, name) == 0) {

			* we have found it *
			ret = filter_delete(i);
			break;
		}
	}
	return ret;
}
*/

static
int filter_toggle_active(const char* name, int type, int active) {
	int i;
	for (i = 0; i < filters.nfilters; i++) {
		if (filters.filter[i].type == type &&
		    strcmp(filters.filter[i].name, name) == 0) {

			filters.filter[i].active = active;
			return 0;
		}
	}
	return -1;
}

int filter_activate(const char* name, int type) {
	return filter_toggle_active(name, type, 1);
}

int filter_deactivate(const char* name, int type) {
	return filter_toggle_active(name, type, 0);
}

int filters_apply(int type, struct transfer_t* t, struct proxyinfo *pi) {
	filter_func_t f;
	int ret, i;

	for (i = 0; i < filters.nfilters; i++) {
		if (filters.filter[i].type == type &&
		    filters.filter[i].active) {

			f = filters.filter[i].filter_func;

			ret = f(t, pi, &filters.filter[i]);
			if (ret < 0) {
				return ret;
			}
		}
	}
	return 0;
}

int filter_limit_bandwidth(struct transfer_t* t,
			   struct proxyinfo* __unused__,
			   struct filter_t* myself) {
	static float throughput;
	static unsigned long int transferred;
	static time_t start;
	time_t done, delay;

	ASSUME(t != NULL);

	/* reset after each transfer, "static" sets the variable to 0 for the
	 * first transfer */
	if (t->status == TRANSFER_STATUS_ERROR
				|| t->status == TRANSFER_STATUS_FINISHED) {
		throughput = 0.0;
		return 0;
	}
	/* no throughput limitations for this transfer */
	if (throughput < 0.0) {
		return 0;
	}
	/* see if there are throughput limitations for this transfer */
	if (throughput == 0.0) {
		/* check for limits in this direction first. Then check for the
		 * general limit */
		if (myself->type == FILTER_TYPE_UPSTREAM) {
			throughput = config_get_foption("throughput-up", -1.0);
		} else {
			throughput = config_get_foption("throughput-down", -1.0);
		}
		/* if throughput == -1.0 ... actually */
		if (throughput < 0.0) {
			throughput = config_get_foption("throughput", -1.0);
		}

		/* if throughput == -1.0 ... actually */
		if (throughput < 0.0) {
			/* no... */
			return 0;
		} else {
			/* yes, init ... */
			start = time(NULL);
			transferred = 0;
		}
	}
	done = time(NULL);
	delay = (transferred / 1024.0) / throughput;
	delay += start;
	delay -= done;
	/* debug...
	jlog(9, "start: %d, done: %d, transferred: %d, throughput: %f",
					start, done, transferred, throughput);
	jlog(9, "delay: %d", delay);
	*/
	if (delay > 0) {
		sleep(delay);
	}
	/* this is not done before because we haven't actually transmitted this
	 * data. We're a filter in between of the transfer ... */
	transferred += t->bufsize;
	return 0;
}


