/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <syslog.h>
#include <stdarg.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include "jproxygw.h"

extern int chlds_exited;

char* errstr = 0;

int register_pid(pid_t pid,
		unsigned long int from_ip,
		unsigned long int proxy_ip,
		unsigned int proxy_port,
		time_t start_time,
		struct proxyinfo* pi) {

	struct connliststruct *cls, *tmp;

	cls = (struct connliststruct*) malloc(sizeof(struct connliststruct));
	enough_mem(cls);

	cls->next = (struct connliststruct*) 0;
	cls->pid = pid;
	cls->from_ip = from_ip;
	cls->proxy_ip = proxy_ip;
	cls->proxy_port = proxy_port;
	cls->start_time = start_time;
	jlog(9, "Adding pid %d", cls->pid);

	if (!pi->si->connection.connected_clients) {
		/* first entry */
		pi->si->connection.connected_clients = cls;
	} else {
		tmp = pi->si->connection.connected_clients;
		while (tmp->next) {
			tmp = tmp->next;
		}
		tmp->next = cls;
	}

	return 0;
}


int unregister_pid(pid_t pid, struct proxyinfo* pi) {

	struct connliststruct* cls, *prev = 0;

	cls = pi->si->connection.connected_clients;

	if (!cls) {
		/* empty! Error! */
		return 1;
	}

	while (cls) {
		jlog(8, "checking %d against %d", cls->pid, pid);
		if (cls->pid == pid) {
			/* found */
			/* delete it */
			/* case 1: prev is set */
			if (prev) {
				prev->next = cls->next;
			}

			/* case 2: prev not set */
			  else {
				/* first element */
				if (!cls->next) {
					/* only one element */
					pi->si->connection.connected_clients
						= (struct connliststruct*) 0;
				} else {
					/* nothing special, it's just the
					 * first element */
					pi->si->connection.connected_clients
						= cls->next;
				}
			}
			config_counter_decrease(cls->from_ip,
						cls->proxy_ip,
						cls->proxy_port,
						cls->start_time,
						pi->si->servertype,
						&(pi->si->hostcache));
			/* remove that element */
			free(cls);
			return 0;
		}
		prev = cls;
		cls = cls->next;
	}
	/* not found */
	return 1;
}


int set_conffilename(const char* arg, struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;
	const int increment = 1;
	int pathsize = 1;
	char* ret;
	char* path;
	char* fname;

	if (!arg || strlen(arg) > PATH_MAX) {
		free(si->conffilename);
		si->conffilename = (char*) 0;
		return -1;
	}

	path = (char*) malloc(pathsize);
	enough_mem(path);
	free(si->conffilename);
	while (!(ret = getcwd(path, pathsize))) {
		pathsize += increment;
		if (pathsize > PATH_MAX) {
			si->conffilename = (char*) 0;
			free(path);
			return -1;
		}
		path = (char*) realloc(path, pathsize);
	}

	fname = (char*) malloc(PATH_MAX + 1);
	si->conffilename = rel2abs(arg, path, fname, PATH_MAX);
	if (!si->conffilename) {
		free(path); free(fname);
		return -1;
	} else {
		/* reduce allocated memory */
		si->conffilename = realloc(fname, strlen(fname) + 1);
		enough_mem(si->conffilename);
		free(path);
	}
	jlog(9, "Conffile set to %s", si->conffilename);
	return 0;
}


int save_runasuser_uid(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;

	const char* option;
	struct passwd* pws;
	struct group* gs;

	option = config_get_option("runasuser");
	if (option) {
		/* we might be in a chroot-environment and not be able to look
		 * up the uid/gid values. That's why we have saved them... So
		 * before changing our cached values we check if we can look up
		 * the data. If we can't, it's fine and we use the cached
		 * values */
		if (si->chrooted) {
			pws = getpwnam(option);
			if (!pws) {
				/* we can't look it up... let's use the cached
				 * values */
				jlog(7, "we are in a chrooted environment and "
					"we can't look up new User IDs. So we "
					"continue to use the old values");
				return 0;
			}
		}
		if (si->runasuser.username) { free(si->runasuser.username); }
		si->runasuser.username = strdup(option);
		enough_mem(si->runasuser.username);
		si->runasuser.username = trim(si->runasuser.username);
		pws = getpwnam(si->runasuser.username);
		if (!pws) {
			jlog(3, "getpwnam(2) could not look up user %s",
				si->runasuser.username);
			fprintf(stderr, "Could not look up the user name.\n");
			return -1;
		}
		si->runasuser.uid = pws->pw_uid;
	} else {
		si->runasuser.uid = getuid();
	}

	option = config_get_option("runasgroup");
	if (option) {
		if (si->runasuser.groupname) { free(si->runasuser.groupname); }
		si->runasuser.groupname = strdup(option);
		enough_mem(si->runasuser.groupname);
		si->runasuser.groupname = trim(si->runasuser.groupname);
		gs = getgrnam(si->runasuser.groupname);
		if (!gs) {
			jlog(3, "getpwnam(2) could not look up group %s",
						si->runasuser.groupname);
			fprintf(stderr, "Could not look up the group name.\n");
			return -1;
		}
		si->runasuser.gid = gs->gr_gid;
	} else {
		/* fallback value */
		si->runasuser.gid = getgid();
		/* find the default group for the user */
		if ((option = config_get_option("runasuser"))) {
			pws = getpwnam(option);
			/* if it was NULL we would have bailed out before */
			ASSUME(pws != NULL);
			si->runasuser.gid = pws->pw_gid;
			gs = getgrgid(pws->pw_gid);
			if (!gs) {
				fprintf(stderr, "Could not look up default "
						"group for user %s. gid %d "
						"does not exist.\n",
						option, pws->pw_gid);
				return -1;
			}
			ASSUME(gs->gr_name != NULL);
			si->runasuser.groupname = strdup(gs->gr_name);
			enough_mem(si->runasuser.groupname);
		}
	}
	return 0;
}

void set_errstr(const char* s) {
	if (errstr) {
		free(errstr);
	}
	errstr = strdup(s);
}

const char* get_errstr(void) {
	if (errstr) {
		return errstr;
	} else {
		return "No detailed error information available... :-(";
	}
}

void free_errstr(void) {
	if (errstr) {
		free(errstr);
		errstr = (char*) 0;
	}
}

/* get the status of the child and unregister it */
/* this function is the signal handler */
void childterm (int signo) {
	chlds_exited++;
}

int get_chld_pid(struct proxyinfo* pi) {
	int ret;
	int status;
	pid_t pid;

	while ((pid = waitpid (-1, &status, WNOHANG)) > 0) {
		jlog(9, "A child exited. Pid: %d", pid);
		jlog(9, "unregistering pid ...");
		ret = unregister_pid(pid, pi);
		if (ret) {
			jlog(3, "Error unregistering pid");
		} else {
			jlog(9, "unregistered");
		}
	}
	/* XXX race condition ? */
	chlds_exited = 0;
	return 0;
}

