/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "types.h"
#include "util.h"
#include "socketio.h"
/* XXX for quotstrtok - but I don't like this include */
#include "config_header.h"
#include <stdarg.h> /* for va_list */

/* functions from config.c */
const struct hostent_list* config_forward_lookup(struct hostent_list** hl,
					const char* name);
const struct hostent_list* config_reverse_lookup(struct hostent_list** hl,
					unsigned long int ip);

/* from util.c */
void enough_mem(const void*);

/* from log.c */
extern void jlog(int, const char *fmt, ...);


/* --------begin struct ilist-------------- */

struct ilist_t* ilist_init(int val) {
	struct ilist_t* i;

	i = (struct ilist_t*) malloc(sizeof(struct ilist_t));
	enough_mem(i);

	i -> value = val;
	i -> next = (struct ilist_t*) 0;

	return i;
}

void ilist_destroy(struct ilist_t* il) {
	if (! il) { return; }
	ilist_destroy(il->next);
	free(il);
}

struct ilist_t* ilist_push(struct ilist_t* il, int i) {
	if (!il) {
		return il;
	}
	while(il -> next) {
		il = il -> next;
	}
	il -> next = ilist_init(i);
	return il->next;
}

int ilist_pop(struct ilist_t* il) {
	struct ilist_t* prev = (struct ilist_t*) 0;
	int i;

	while(il -> next) {
		prev = il;
		il = il -> next;
	}
	i = il->value;
	if (prev) {
		free(il);
		prev->next = (struct ilist_t*) 0;
	}
	return i;
}

int ilist_empty(struct ilist_t* il) {
	if (!il) {
		return 1;
	}
	return (il->next == (struct ilist_t*) 0);
}

struct ilist_t* ilist_clone(const struct ilist_t* il) {
	struct ilist_t* new;

	if (!il) {
		return (struct ilist_t*) 0;
	}
	new = (struct ilist_t*) malloc(sizeof(struct ilist_t));
	new->value = il->value;
	new->next = ilist_clone(il->next);

	return new;
}

/* --------end struct ilist-------------- */



/* --------begin struct ullist-------------- */

struct ullist_t* ullist_init(unsigned long int val) {
	struct ullist_t* ul;

	ul = (struct ullist_t*) malloc(sizeof(struct ullist_t));
	enough_mem(ul);

	ul -> value = val;
	ul -> next = (struct ullist_t*) 0;

	return ul;
}

void ullist_destroy(struct ullist_t* ul) {
	if (! ul) { return; }
	ullist_destroy(ul->next);
	free(ul);
}

struct ullist_t* ullist_push(struct ullist_t* ul, unsigned long int ulval) {
	if (!ul) {
		return ul;
	}
	while(ul -> next) {
		ul = ul -> next;
	}
	ul -> next = ullist_init(ulval);
	return ul->next;
}

int ullist_pop(struct ullist_t* ul) {
	struct ullist_t* prev = (struct ullist_t*) 0;
	unsigned long int val;

	while(ul -> next) {
		prev = ul;
		ul = ul -> next;
	}
	val = ul->value;
	if (prev) {
		free(ul);
		prev->next = (struct ullist_t*) 0;
	}
	return val;
}

int ullist_empty(struct ullist_t* ul) {
	if (!ul) {
		return 1;
	}
	return (ul->next == (struct ullist_t*) 0);
}

const char* gethostentip(const char* iplist) {
	static char ipbuf[16];
	snprintf(ipbuf, 16, "%d.%d.%d.%d",
		(unsigned char) iplist[0],
		(unsigned char) iplist[1],
		(unsigned char) iplist[2],
		(unsigned char) iplist[3]
	);
	return ipbuf;
}

struct ullist_t* ullist_conv_ips(char** ip_list) {
	int i = 0;
	struct ullist_t* base = (struct ullist_t*) 0,
			*cur = (struct ullist_t*) 0;
	unsigned long int ip;

	while (ip_list[i]) {
		ip = inet_addr(gethostentip(ip_list[i]));
		if (i == 0) {
			base = cur = ullist_init(ip);
		} else {
			cur = ullist_push(cur, ip);
		}
		i++;
	}
	return base;
}

/* --------end struct ullist-------------- */


/* --------begin struct slist-------------- */

struct slist_t* slist_init(char* val) {
	struct slist_t* s;

	s = (struct slist_t*) malloc(sizeof(struct slist_t));
	enough_mem(s);

	s -> value = val;
	s -> next = (struct slist_t*) 0;

	return s;
}


/* slist_cinit  copies */
struct slist_t* slist_cinit(const char* val) {
	struct slist_t* s;

	s = slist_init((char*) 0);
	s -> value = strdup(val);
	enough_mem(s->value);
	return s;
}

void slist_destroy_free(struct slist_t* sl) {
	if (! sl) { return; }
	slist_destroy_free(sl->next);
	free(sl->value);
	free(sl);
}

struct slist_t* slist_push(struct slist_t* sl, char* s) {
	if (!sl) {
		return sl;
	}
	if (!sl->value) {
		/* empty list */
		sl->value = s;
		return sl;
	}
	while(sl -> next) {
		sl = sl -> next;
	}
	sl -> next = slist_init(s);
	return sl->next;
}

/* copy before pushing */
struct slist_t* slist_cpush(struct slist_t* sl, const char* s) {
	char* a = strdup(s);
	enough_mem(a);

	return slist_push(sl, a);
}

struct slist_t* slist_append(struct slist_t* a, struct slist_t* b) {
	struct slist_t* orig = a;

	if (!a) {
		return b;
	}
	while (a->next) {
		a = a->next;
	}
	a->next = b;

	return orig;
}

char* slist_get(struct slist_t* sl) {
	return sl->value;
}

/* the caller has to free the base element for himself */
char* slist_pop(struct slist_t* sl) {
	struct slist_t* prev = (struct slist_t*) 0;
	char *s;

	if (!sl) {
		return (char*) 0;
	}
	while(sl -> next) {
		prev = sl;
		sl = sl -> next;
	}
	s = sl->value;
	if (prev) {
		/* don't free value */
		free(sl);
		prev->next = (struct slist_t*) 0;
	} else {
		/* first element */
		sl->value = (char*) 0;
	}
	return s;
}


int slist_empty(struct slist_t* sl) {
	if (!sl) {
		return 1;
	}
	return (sl->next == (struct slist_t*) 0);
}

struct slist_t* slist_conv_dpointer(char** dpointer) {
	int i = 0;
	struct slist_t* sl;
	char* s;

	if (dpointer[0]) {
		s = strdup(dpointer[0]);
		enough_mem(s);
		sl = slist_init(s);
	} else {
		return (struct slist_t*) 0;
	}

	i = 1;
	while (dpointer[i]) {
		s = strdup(dpointer[i]);
		enough_mem(s);
		slist_push(sl, s);
		i++;
	}
	return sl;
}

struct slist_t* slist_clone(const struct slist_t* sl) {
	struct slist_t* slret, *slcur;

	if ( ! sl ) {
		return (struct slist_t*) 0;
	}
	slret = slcur = slist_cinit(sl->value);
	sl = sl->next;
	while (sl) {
		slist_cpush(slcur, sl->value);
		slcur = slcur->next;
		sl    = sl->next;
	}
	return slret;
}

struct slist_t* slist_reverse(struct slist_t* sl) {
	/*   a -> b -> c
	 *
	 * becomes
	 *
	 *   c -> b -> a
	 *
	 **/

	struct slist_t* head, *curr, *change;

	if (! sl) {
		return (struct slist_t*) 0;
	}

	if (! sl->next) {
		/* only one element */
		return sl;
	}

	curr = sl;
	change = sl->next;
	head = sl->next;

	curr->next = (struct slist_t*) 0;

	while ( head->next ) {

		head = head->next;
		change->next = curr;

		curr = change;
		change = head;
	};

	/* head has reached the end */
	head->next = curr;

	return head;
}

char* slist_concat(struct slist_t* sl) {
	char* str;
	size_t length = 0;

	if (!sl) {
		return (char*) 0;
	}

	str = malloc(1);
	str[0] = '\0';
	enough_mem(str);

	do {
		length += strlen(sl->value);
		str = realloc(str, length + 1);
		enough_mem(str);
		strcat(str, sl->value);
	} while ((sl = sl->next));

	return str;
}

int slist_case_contains(const struct slist_t* haystack, const char* needle) {
	while (haystack) {
		if (strcasecmp(haystack->value, needle) == 0) {
			return 1;
		}
		haystack = haystack->next;
	}
	return 0;
}

int slist_count(const struct slist_t* haystack) {
	int counter = 0;
	while (haystack) {
		counter++;
		haystack = haystack->next;
	}
	return counter;
}

/* --------end struct slist-------------- */

/* --------begin struct optionlist-------------- */

struct option_t* optionlist_init(const char* key, const char* value) {
	struct option_t* o;

	o = (struct option_t*) malloc(sizeof(struct option_t));
	enough_mem(o);

	o->key  = strdup(key);
	enough_mem(o->key);
	o->value = strdup(value);
	enough_mem(o->value);
	o->next = (struct option_t*) 0;

	return o;
}

struct option_t* optionlist_push(struct option_t* o,
					const char* key,
					const char* value) {

	if (! o) {
		return (struct option_t*) 0;
	}
	while (o->next) {
		o = o->next;
	}
	o->next  = optionlist_init(key, value);

	return o;
}

struct option_t* optionlist_clone(struct option_t* orig) {
	struct option_t* ret, *trav;
	if ( ! orig ) {
		return (struct option_t*) 0;
	}

	ret = optionlist_init(orig->key, orig->value);
	trav = ret;

	while ((orig = orig->next)) {
		trav->next = optionlist_init(orig->key, orig->value);
		trav = trav->next;
	}
	return ret;
}

struct option_t* optionlist_append(struct option_t* a, struct option_t* b) {
	struct option_t* orig = a;

	if (!a) {
		return b;
	}
	while (a->next) {
		a = a->next;
	}
	a->next = b;

	return orig;
}

void optionlist_destroy(struct option_t* olist) {
	if ( !olist ) { return; }
	optionlist_destroy( olist->next );
	free( olist->key );
	free( olist->value );
	free( olist );
}

void optionlist_delete_key(struct option_t* olist, const char* key) {
	struct option_t* deleted;

	while (olist) {
		if (olist->next) {
			if (strcasecmp(olist->next->key, key) == 0) {
				/* found */
				deleted = olist->next;
				olist->next = olist->next->next;
				deleted->next = (struct option_t*) 0;
				optionlist_destroy(deleted);
			}
		}
		olist = olist->next;
	}
}

const char* option_get_value(const struct option_t* options, const char* key) {
	const char* val = (char*) 0;

	if ( ! key ) {
		return (char*) 0;
	}

	while (key && options) {
		if (options->key && strcasecmp(options->key, key) == 0) {
			val = options->value;
		}
		options = options->next;
	}
	return val;
}

/* --------end struct optionlist-------------- */


/* --------begin struct option_prio_list-------------- */

struct option_prio_t* option_prio_list_init(const char* key,
						const char* value,
						int priority) {
	struct option_prio_t* o;

	o = (struct option_prio_t*) malloc(sizeof(struct option_prio_t));
	enough_mem(o);

	o->key  = strdup(key);
	enough_mem(o->key);
	o->value = strdup(value);
	enough_mem(o->value);
	o->priority = priority;
	o->next = (struct option_prio_t*) 0;

	return o;
}

struct option_prio_t* option_prio_list_push(struct option_prio_t* o,
						const char* key,
						const char* value,
						int priority) {

	if (! o) {
		return (struct option_prio_t*) 0;
	}
	while (o->next) {
		o = o->next;
	}
	o->next  = option_prio_list_init(key, value, priority);

	return o;
}

struct option_prio_t* option_prio_list_clone(struct option_prio_t* orig) {
	struct option_prio_t* ret, *trav;
	if ( ! orig ) {
		return (struct option_prio_t*) 0;
	}

	ret = option_prio_list_init(orig->key, orig->value, orig->priority);
	trav = ret;

	while ((orig = orig->next)) {
		trav->next = option_prio_list_init(orig->key,
							orig->value,
							orig->priority);
		trav = trav->next;
	}
	return ret;
}

void option_prio_list_destroy(struct option_prio_t* olist) {
	if ( !olist ) { return; }
	option_prio_list_destroy( olist->next );
	free( olist->key );
	free( olist->value );
	free( olist );
}

int option_prio_list_count(const struct option_prio_t* olist) {
	int count = 0;
	while (olist) {
		olist = olist->next;
		count++;
	}
	return count;
}

static
int option_priority_compare(const void* a_v, const void* b_v) {
	const struct option_prio_t *a = (const struct option_prio_t*) a_v;
	const struct option_prio_t *b = (const struct option_prio_t*) b_v;

	if (a->priority < b->priority) {
		return -1;
	}
	if (a->priority > b->priority) {
		return 1;
	}
	return 0;
}

struct option_t* option_prio_list_sort(const struct option_prio_t* olist) {
	struct option_t* olist_sorted = (struct option_t*) 0;
	struct option_prio_t* arr;
	int size;
	int i;

	if (! olist) {
		return (struct option_t*) 0;
	}

	/* sort the list */
	size = option_prio_list_count(olist);
	arr = (struct option_prio_t*) malloc(sizeof(struct option_prio_t)
								* size);
	enough_mem(arr);
	for (i = 0; i < size; i++) {
		ASSUME(olist != NULL);
		arr[i].key = olist->key;
		arr[i].value = olist->value;
		arr[i].priority = olist->priority;
		olist = olist->next;
	}

	qsort(&arr[0], size, sizeof(struct option_prio_t),
						option_priority_compare);

	/* init the list manually with the first element */
	olist_sorted = optionlist_init(arr[0].key, arr[0].value);

	for(i = 1; i < size; i++) {
		optionlist_push(olist_sorted, arr[i].key, arr[i].value);
	}

	free(arr);
	return olist_sorted;
}

struct option_prio_t* option_prio_list_append(struct option_prio_t* a,
						struct option_prio_t* b) {
	struct option_prio_t* orig = a;

	if (!a) {
		return b;
	}
	while (a->next) {
		a = a->next;
	}
	a->next = b;

	return orig;
}

/* --------end struct optionlist-------------- */



/* --------begin struct hostent_list-------------- */

struct hostent_list* hostent_init(struct hostent* e,
					unsigned long int ip,
					const char* name) {

	struct hostent_list* h;

	h = (struct hostent_list*) malloc(sizeof(struct hostent_list));
	enough_mem(h);

	if (name) {
		/* jlog(9, "adding %s to host cache", name); */
	} else {
		/* jlog(9, "adding %s to host cache",
				inet_ntoa(*((struct in_addr*) &ip))); */
	}

	h->next = (struct hostent_list*) 0;
	if (name) {
		h->name = strdup(name);
	} else {
		h->name = (char*) 0;
	}
	h->ip = ip;

	if (e) {
		h->aliases_list = slist_conv_dpointer(e->h_aliases);
		h->addr_list = ullist_conv_ips(e->h_addr_list);
		if (e->h_name) {
			if (h->aliases_list) {
				slist_cpush(h->aliases_list, e->h_name);
			} else {
				h->aliases_list = slist_cinit(e->h_name);
			}
		} else {
			h->aliases_list = (struct slist_t*) 0;
			h->addr_list    = (struct ullist_t*) 0;
		}
	} else {
		h->aliases_list = (struct slist_t*) 0;
		h->addr_list    = (struct ullist_t*) 0;
	}
	h->lookup_time = time(NULL);
	return h;
}

struct hostent_list* hostent_push(struct hostent_list* h,
					struct hostent* e,
					unsigned long int ip,
					const char* name) {
	if (!h) {
		return (struct hostent_list*) 0;
	}
	while (h->next) {
		h = h->next;
	}
	h->next = hostent_init(e, ip, name);
	return h;
}

const struct hostent_list* hostent_get(struct hostent_list* h,
				unsigned long int ip,
				const char* name) {

	while (h) {
		if (name && h->name) {
			if (strcasecmp(name, h->name) == 0) {
				return h;
			}
		} else {
			if (ip != -1 && ip == h->ip) {
				return h;
			}
		}
		h = h->next;
	}
	return (struct hostent_list*) 0;
}

const struct slist_t* hostent_get_aliases(struct hostent_list** h,
					unsigned long int ip) {
	const struct hostent_list* hl;

	hl = config_reverse_lookup(h, ip);
	if (!hl) {
		return (struct slist_t*) 0;
	}
	return hl->aliases_list;
}

const struct ullist_t* hostent_get_addr(struct hostent_list** h,
					const char* name) {
	const struct hostent_list* hl;

	hl = config_forward_lookup(h, name);
	if (!hl) {
		return (struct ullist_t*) 0;
	}
	return hl->addr_list;
}

const char* hostent_get_name(struct hostent_list** h,
					unsigned long int ip) {

	const struct slist_t* l = hostent_get_aliases(h, ip);
	if (!l) {
		return (char*) 0;
	}
	return l->value;
}

unsigned long int hostent_get_ip(struct hostent_list** h,
					const char* name) {
	const struct ullist_t* l = hostent_get_addr(h, name);
	if (!name) {
		return UINT_MAX;
	}
	if (!l) {
		/* Maybe name is already an IP ? */
		unsigned long int iptest;
		iptest = inet_addr(name);
		if (iptest != (unsigned long int) UINT_MAX) {
			/* it was an IP */
			return iptest;
		}
		return UINT_MAX;
	}
	return l->value;
}

void hostent_destroy(struct hostent_list* h) {
	if (!h) { return; }
	hostent_destroy(h->next);
	ullist_destroy(h->addr_list);
	slist_destroy_free(h->aliases_list);
	free(h->name);
	free(h);
}

void hostent_delete(struct hostent_list* hl, const struct hostent_list* entry){
	struct hostent_list* delete;
	if (!hl || !entry) {
		return;
	}
	while (hl) {
		if (hl->next == entry) {
			delete = hl->next;
			hl->next = hl->next->next;
			delete->next = (struct hostent_list*) 0;
			hostent_destroy(delete); 
			return;
		} else {
			hl = hl->next;
		}
	}
}

/* --------end struct hostent_list-------------- */

/* --------begin struct hostlist_t-------------- */

struct hostlist_t* hostlist_init() {
	struct hostlist_t* hl;
	hl = (struct hostlist_t*) malloc(sizeof(struct hostlist_t));
	enough_mem(hl);

	hl->next                     = (struct hostlist_t*) 0;
	hl->host.ip.ip               = -1;
	hl->host.ip.netmask          = -1;
	hl->host.name                = (char*) 0;

	return hl;
}


struct hostlist_t* hostlist_push(struct hostlist_t** hl,
				 char* name,
				 unsigned long int ip,
				 unsigned long int netmask) {

	struct hostlist_t* h;
	if (! hl) {
		/* this should NOT happen */
		jlog(3, "hostlist was NULL in hostlist_push()");
		return (struct hostlist_t*) 0;
	}
	h = *hl;
	if (! h) {
		/* new */
		h = hostlist_init();
		*hl = h;
	} else {
		while (h->next) {
			h = h->next;
		}
		h->next = hostlist_init();
		h = h->next;
	}

	/* fill with content */
	h->next = (struct hostlist_t*) 0;
	h->host.ip.ip           = ip;
	h->host.ip.netmask      = netmask;
	h->host.name            = name;

	/* return base address */
	return *hl;
}

struct hostlist_t* hostlist_clone(struct hostlist_t* hl) {
	struct hostlist_t* new;

	if ( ! hl ) {
		return (struct hostlist_t*) 0;
	}

	new = (struct hostlist_t*) malloc(sizeof(struct hostlist_t));

	new->host.ip = hl->host.ip;
	if (hl->host.name) {
		new->host.name = strdup(hl->host.name);
		enough_mem(new->host.name);
	} else {
		new->host.name = (char*) 0;
	}
	new->next = hostlist_clone(hl->next);
	return new;
}

struct hostlist_t* hostlist_ip_push(struct hostlist_t** hl,
				    unsigned long int ip,
				    unsigned long int netmask) {

	return hostlist_push(hl, (char*) 0, ip, netmask);
}

struct hostlist_t* hostlist_name_push(struct hostlist_t** hl,
				      char* name) {

	return hostlist_push(hl, name, -1, -1);
}

/* --------end struct hostlist_t---------------- */


/* --------begin struct portrangestruct-------------- */

struct portrangestruct* portrangestruct_port2portrange(unsigned int port) {
	static struct portrangestruct prs;
	prs.next = (struct portrangestruct*) 0;
	prs.startport = prs.endport = port;
	return &prs;
}

unsigned int portrangestruct_count(const struct portrangestruct* prs) {
	if ( ! prs ) {
		return 0;
	}
	return (prs->endport - prs->startport + 1) +
		portrangestruct_count(prs->next);
}

struct portrangestruct* portrangestruct_clone(struct portrangestruct* prs) {
	struct portrangestruct* new;
	if ( ! prs ) {
		return (struct portrangestruct*) 0;
	}

	new = (struct portrangestruct*) malloc(sizeof(struct portrangestruct));
	enough_mem(new);

	new->startport = prs->startport;
	new->endport = prs->endport;

	new->next = portrangestruct_clone(prs->next);

	return new;
}

/* --------end struct portrangestruct---------------- */

/* --------start struct ip--------------------------- */

struct ip_t ip_parse_from_char(const char* str) {
	/* expand to parse 123.123.123.123/255.255.0.0 for example */
	/* XXX there is sth. in util.c that does this! */
	struct ip_t ip;
	struct in_addr addr;
	int ret;

	ret = inet_aton(str, &addr);
	if (ret == 0) {
		/* invalid ip */
		ip.error = 1;
	} else {
		ip.error = 0;
		ip.ip = addr.s_addr;
		ip.netmask = inet_addr("255.255.255.255");
	}
	return ip;
}

struct ip_t ip_from_s_addr(unsigned int s_addr) {
	struct ip_t ip;

	ip.error = 0;
	ip.ip = s_addr;
	ip.netmask = inet_addr("255.255.255.255");
	return ip;
}

char* ip_to_mchar(struct ip_t ip) {
	char* buffer;
	if (ip.error) {
		buffer = strdup("255.255.255.255");
		enough_mem(buffer);
	} else {
		int size = 3*4 + 3 + 1;
		char* tmp = inet_ntoa(*((struct in_addr*) &ip.ip));
		buffer = malloc(size);
		enough_mem(buffer);
		ASSUME(size >= strlen(tmp) + 1);
		strcpy(buffer, tmp);
	}
	return buffer;
}

/* --------end struct ip----------------------------- */

/* --------start struct address---------------------- */

/* the error attribute is special. There are several of them because there is
 * one for the ip, one for the host and one for the address. Not all of them
 * have to be set correctly, i.e. if address.error is set, it doesn't matter if
 * address.ip.error is set or cleared. However if address.error is set, it is
 * guaranteed that we don't have to free address.host.name for example.
 */

/* XXX error checking */
char* address_ip_mchar(struct address_t address) {
	return mconv_ip2char(address.host.ip.ip);
}

unsigned int address_port(struct address_t address) {
	return address.port;
}

unsigned int address_port_hs(struct address_t address) {
	/* XXX verify */
	return ntohs(address.port);
}

unsigned int address_port_ns(struct address_t address) {
	/* XXX verify */
	return htons(address.port);
}

/* return a textual representation of the hostname. If possible, prefer the
 * name, if it is not set, convert the ip to a char */
char* address_host_pref_name_mchar(struct address_t address) {
	char* result;
	if (address.error == 1) {
		return (char*) 0;
	}
	if (address.host.error == 1) {
		return (char*) 0;
	}
	if (address.host.name) {
		/* ok we can get the hostname */
		result = strdup(address.host.name);
		enough_mem(result);
	} else {
		ASSUME(address.host.ip.error == 0);
		result = address_ip_mchar(address);
	}
	return result;
}

char* address_ip_c_port_mchar(struct address_t address) {
	char* s_ip, *s_port;
	char* result;
	unsigned int port;
	s_ip = mconv_ip2char(address.host.ip.ip);
	port = address_port_hs(address);
	s_port = mconv_uint2char(port);

	result = char_enclose(s_ip, ":", s_port);

	free(s_ip);
	free(s_port);
	return result;
}

struct address_t address_parse_host_c_port(const char* str) {
	struct address_t addr;
	char* hoststr;
	char* portstr;
	int offset = 0;

	addr.error = 0;
	if (!str) {
		addr.error = 1;
		return addr;
	}
	hoststr = quotstrtok(str, ":, \t", &offset);
	if (!hoststr) {
		addr.error = 1;
		return addr;
	}
	if (strlen(hoststr) == 0) {
		free(hoststr);
		hoststr = (char*) 0;
		addr.error = 1;
		return addr;
	}

	addr.host.ip = ip_parse_from_char(hoststr);
	if (addr.host.ip.error) {
		addr.host.name = hoststr;
		addr.host.error = 0;
	}
	portstr = quotstrtok(str, ":, \t", &offset);
	addr.port = conv_char2long(portstr, 0);
	if (addr.port == 0) {
		free(addr.host.name);
		addr.host.name = (char*) 0;
		addr.error = 1;
	}
	free(portstr);
	return addr;
}

void address_free(struct address_t address) {
	if (address.error) {
		return;
	}
	if (address.host.error) {
		return;
	}
	if (address.host.name) {
		free(address.host.name);
	}
}

/* ----------end struct address---------------------- */


/* --------start struct messagebuffer---------------- */
void message_buffer_initialize(struct messagebuffer* msgbuf) {
	msgbuf->fd = -1;
	msgbuf->code = -1;
	msgbuf->msg = (char*) 0;
	msgbuf->next = (struct messagebuffer*) 0;
}

void message_buffer_reset(struct messagebuffer* msgbuf) {
	if (msgbuf->fd == -1) {
		/* it's already in the basic state */
		return;
	}
	if (msgbuf->next) {
		message_buffer_reset(msgbuf->next);
		free(msgbuf->next);
	}
	free(msgbuf->msg);
	message_buffer_initialize(msgbuf);
}

void message_buffer_add_nomalloc(struct messagebuffer* msgbuf,
			int fd, int code, char* msg) {
	if (msgbuf->fd == -1) {
		/* ok, use this */
		msgbuf->fd = fd;
		msgbuf->code = code;
		msgbuf->msg = msg;
		return;
	}
	while (msgbuf->next) {
		msgbuf = msgbuf->next;
	}
	msgbuf->next = (struct messagebuffer*)
			malloc(sizeof(struct messagebuffer));
	enough_mem(msgbuf->next);
	message_buffer_initialize(msgbuf->next);
	message_buffer_add_nomalloc(msgbuf->next, fd, code, msg);
}

void message_buffer_add(struct messagebuffer* msgbuf,
		       int fd, int code, const char* msg) {
	char* malloc_msg = strdup(msg);
	enough_mem(malloc_msg);
	message_buffer_add_nomalloc(msgbuf, fd, code, malloc_msg);
}

void message_buffer_addf(struct messagebuffer* msgbuf,
			 int fd, int code, const char* fmt, ...) {
	/* XXX static ? */
	char buffer[200];
	va_list vl;

	va_start(vl, fmt);
	vsnprintf(buffer, sizeof(buffer) - 1, fmt, vl);
	va_end(vl);
	message_buffer_add(msgbuf, fd, code, buffer);
}

char* message_buffer_getmsg_m(const struct messagebuffer* msgbuf,
				int fd,
				int *code) {
	/* determine the code for the whole message */
	int final_code = 500;
	const struct messagebuffer* msgbuf_it = msgbuf;
	struct slist_t* slist = (struct slist_t*) 0;
	char* last, *new, *msg, *msg_no_codes;

	/* we get the last code that is != 0 for one fd */
	do {
		if (msgbuf_it->fd == fd && msgbuf_it->code) {
			final_code = msgbuf_it->code;
		}
	} while ((msgbuf_it = msgbuf_it->next));
	*code = final_code;

	/* get all our messages and put them in a buffer */
	msgbuf_it = msgbuf;
	do {
		if (msgbuf_it->fd == fd) {
			new = prependcode(msgbuf_it->msg, final_code, 0);
			if (slist) {
				slist_push(slist, new);
			} else {
				slist = slist_init(new);
			}
		}
	} while ((msgbuf_it = msgbuf_it->next));

	/* get the last one out again and get a terminated version */
	last = slist_pop(slist);
	if (!last) {
		/* there are no messages - nothing to free */
		return (char*) 0;
	}
	ASSUME(strlen(last) >= 4);
	/* shift over the old code and the dash */
	msg_no_codes = strip_respcodes(last);

	/* strip_respcodes shouldn't return a NULL value if the input is not
	 * null (which is obviously not the case) */
	ASSUME(msg_no_codes != NULL);
	new = prependcode(msg_no_codes, final_code, 1);
	free(msg_no_codes);
	slist_push(slist, new);
	free(last);

	/* now we have to concatenate all our strings in slist */
	msg = slist_concat(slist);

	/* and free the list */
	while ((last = slist_pop(slist))) {
		free(last);
	}

	return msg;
}

int message_buffer_say_reset(struct messagebuffer* msgbuf, int fd, int* code) {
	char* msg;
	int ret;
	
	msg = message_buffer_getmsg_m(msgbuf, fd, code);
	/* XXX we reset it for ALL fds... - good or bad? */
	message_buffer_reset(msgbuf);
	ret = say(fd, msg);
	free(msg);
	return ret;
}

/* --------end struct messagebuffer----------------- */

